/**
 * Linux Implementation for the Prophet Address Allocation method
 * see http://www.cse.msu.edu/~zhouhon1/doc/ProphetAddressAlloc.PDF
 * 
 * Copyright (C) 2004 Andreas Gohr <andi@splitbrain.org>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 */

#include "prophet.h"
#include <stdlib.h>
#include <sys/select.h>
#include <sys/time.h>
#include <libnet.h>
#include <pcap.h>

// ---------- start stuff ------------

int main(int argc, char *argv[]) {
    libnet_t *l;
    int i=0;
    u_long* hw;

    //get device from commandline
    if(argc != 2){
        usage();
        exit(1);
    }else{
      _device = argv[1];
    }
    
    //get handle to libnet
    _libnet = get_libnethandle();
    //initialize random generator with harware address
    hw = (u_long*) libnet_get_hwaddr(_libnet);
    srand((int)*hw);

    //do the magic
    if(!request_address()){
        //we're 1st node
        bootstrap();
    }
    wait_for_requests();
}

void usage(){
  printf("  Prophet Address Allocation daemon\n\n");
  printf("  Usage  : ./prophet <device>\n");
  printf("  Example: ./prophet eth1\n");
}

/**
 * Assign the given IP address to the device
 *
 * used http://groups.google.de/groups?th=f945bc2aeb2cd25 as example
 */
int setIP(u_long ip){
    struct ifreq ifr;
    int sd;
    struct sockaddr_in *addr;
    struct in_addr sa;

    //which device to work on:
    strcpy(ifr.ifr_name,_device);

    //get socket to kernel interface
    if( (sd=socket(PF_INET, SOCK_STREAM, 0)) == -1 ) {
        printf( "Getting socket: %s\n", strerror(errno) );
        exit(-1);
    }

    //put IP into structure
    sa.s_addr = ip;

    //set new IP Address
    addr = (struct sockaddr_in *)(&(ifr.ifr_addr));
    memcpy(&addr->sin_addr, &sa, sizeof(struct in_addr));
    addr->sin_family = AF_INET;

    //make system call
    if( ioctl(sd, SIOCSIFADDR, &ifr) == -1 ) {
        close(sd);
        printf("Setting device address: %s\n",
                strerror(errno));
        printf("ifr address = %s",
                inet_ntoa(addr->sin_addr) );
        exit(-1);
    }
    close(sd);

    printf("Set %s's address to %s\n",_device,
           inet_ntoa(addr->sin_addr));
    return TRUE;
}

/**
 * bootstraps the manet if this is the first node.
 * Initializes the state, generates a Network ID and assigns a
 * random IP address
 */
void bootstrap(){
    int i;

    //setup random ip
    setIP(randIP());
    //initialize state
    for (i=0; i<K; i++){
        _prophet.state[i] = 1;
    }
    _prophet.index = 0;
    //create random network id
    _prophet.nid = 1+rand();
}

/**
 * calculates a new IP-Address by using the current state
 */
u_long calcIP(){
    u_long ip;        // this is my address
    u_long address=1; // holds the new address
    int i = 0;

    ip = libnet_get_ipaddr4(_libnet);
    
    //calc new address
    for(i=0; i<K; i++){
        address *= _prime[i]^_prophet.state[i];
				address %= RANGE;
    }
    address += ip;
    address %= RANGE;
    address += 1;
    address += BASENET;
    address  = ntohl(address); //convert byte order
    
    //increase state at index_pos
    _prophet.state[_prophet.index]++;

    printf("generated new address %s\n",
     libnet_addr2name4(address,LIBNET_DONT_RESOLVE));
    return address;
}

/**
 * Generates a random IP-Address
 */
u_long randIP(){
    u_long ip;
    ip = (1+(rand() % RANGE))+BASENET;
    ip = ntohl(ip); //convert byte order
    
    printf("generated pseudo random address %s\n",
           libnet_addr2name4(ip,LIBNET_DONT_RESOLVE));
    return ip;
}

/**
 * Opens the networkdevice to capture ARP packets
 * and returns a handle to the capture device
 */
pcap_t* get_pcaphandle(){
    char errbuf[PCAP_ERRBUF_SIZE];
    pcap_t* descr;
    struct bpf_program filter;

    //open device for capture
    descr = pcap_open_live(_device, // device
                           BUFSIZ,  // snaplength
                           0,       // promiscuous mode
                           -1,      // timout: does not work as expected
                           errbuf); // errors
    if(descr == NULL){
        printf("pcap_open_live(): %s\n",errbuf);
        exit(1);
    }

    // compile filter to only see arp packages
    if(pcap_compile(descr,&filter,"arp",0,0) == -1)
    { fprintf(stderr,"Error calling pcap_compile\n"); exit(1); }
    // set the compiled program as the filter
    if(pcap_setfilter(descr,&filter) == -1)
    { fprintf(stderr,"Error setting filter\n"); exit(1); }
    // free mem for compiled filter
    pcap_freecode(&filter);

    return descr;
}

/**
 * Waits for incoming ARP requests and answers them
 */
int wait_for_requests(){
    pcap_t* pcap;

    printf("Waiting for requests...\n");
    pcap = get_pcaphandle();
    pcap_loop(pcap,-1,check_request,NULL);
    return TRUE;
}

/**
 * Requests an IP Address from its neighbors.
 *
 * returns true if request was fulfilled
 */
int request_address(){
    int try = 0;
    pcap_t* pcap;
    //the following are needed for using select()
    int pcap_fd;
    fd_set pcap_set;
    struct timeval tv;
    int retval;
    const u_char *packet;
    struct pcap_pkthdr hdr;

    while (try < RETRY){
        printf("Sending request... (Try %d)\n",try+1);
        pcap = get_pcaphandle();            // open capture device
        pcap_fd = pcap_fileno(pcap);        // get fildescriptor
        FD_ZERO(&pcap_set);                 // prepare set
        FD_SET(pcap_fd, &pcap_set);         // prepare timer
        tv.tv_sec = 3;                      // wait til next try
        tv.tv_usec = 0;

        //send request here
        send_ARP_request();

        //wait for packets (this loop works only on linux)
        while (tv.tv_sec > 0 || tv.tv_usec >0){
            retval = select(pcap_fd+1, &pcap_set, NULL, NULL, &tv);
            if(retval == -1){
                printf("select() failed\n");
                exit(1);
            }
            if(retval){
                packet = pcap_next(pcap,&hdr);
                //if reply found close pcap and return
                if(check_reply(packet,&hdr)){
                    pcap_close(pcap);
                    return TRUE;
                }
            }
        }
        pcap_close(pcap);        // close capture device
        try++;
    }
    printf("Got no answer to my requests - No life in universe?\n");
    return FALSE;
}

/**
 * Checks if the given packet is a reply and configures the network
 * accordingly
 */
int check_reply(const u_char *packet, struct pcap_pkthdr *hdr){
    struct ether_header *ether;  // defined in net/ethernet.h
    struct sniff_arp *arp;       // defined in prophet.h
    struct prophet_states *data; // the sent states
    u_long *src_ip;
    u_long *dst_ip;
    int size_ethernet = sizeof(struct ether_header);
    int size_arp      = sizeof(struct sniff_arp);

    //get packet into structs
    ether = (struct ether_header*)(packet);
    arp   = (struct sniff_arp*)(packet + size_ethernet);
    data  = (struct prophet_states*)(packet + size_ethernet + size_arp);

    // check opcode
    if(ntohs(arp->arp_op) != ARP_PROPHET_REP){
        return FALSE;
    }

    src_ip = (u_long*) arp->arp_spa;
    dst_ip = (u_long*) arp->arp_tpa;

    //setup ip and states
    printf("Got reply from %s - I shall use %s\n",
            libnet_addr2name4(*src_ip,LIBNET_DONT_RESOLVE),
            libnet_addr2name4(*dst_ip,LIBNET_DONT_RESOLVE));

    memcpy(&_prophet,data,sizeof(struct prophet_states));
    setIP(*dst_ip);

    //increase index
		_prophet.index++;
    _prophet.index %= K;

    return TRUE;
}

/**
 * Handles Requests be answering them
 */
void check_request(u_char *user,const struct pcap_pkthdr* pkthdr,
                   const u_char* packet){
    struct ether_header *ether;  // defined in net/ethernet.h
    struct sniff_arp *arp;       // defined in prophet.h
    int size_ethernet = sizeof(struct ether_header);
    int size_arp      = sizeof(struct sniff_arp);

    //get packet into structs
    ether = (struct ether_header*)(packet);
    arp   = (struct sniff_arp*)(packet + size_ethernet);
    
    // check opcode
    if(ntohs(arp->arp_op) != ARP_PROPHET_REQ){
        return;
    }

    printf("Got request from %s\n", ether_ntoa(arp->arp_sha) );
    send_ARP_reply(arp->arp_sha);
}

/**
 * Returns a handle to the libnet interface
 */
libnet_t* get_libnethandle(){
    libnet_t *l;
    char errbuf[LIBNET_ERRBUF_SIZE];

    //Initialize the library.
    l = libnet_init( LIBNET_LINK_ADV, _device, errbuf);

    if (l == NULL){
        fprintf(stderr, "libnet_init() failed: %s", errbuf);
        exit(EXIT_FAILURE);
    }
    return l;
}

/**
 * sends an ARP request
 */
int send_ARP_request(){
    u_char* my_hw = (u_char*) libnet_get_hwaddr(_libnet);
    return send_ARP(ARP_PROPHET_REQ,my_hw,_ip_nil,_hw_broadcast,_ip_nil);
}

/**
 * sends an ARP reply to the given Hardware address
 */
int send_ARP_reply(u_char* dst_hw){
    u_char* my_hw;
    u_long my_ip;
    u_long newip;

    my_hw = (u_char*) libnet_get_hwaddr(_libnet);
    my_ip = libnet_get_ipaddr4(_libnet);
    newip = calcIP();
    send_ARP(ARP_PROPHET_REP,my_hw,(u_char*) &my_ip,dst_hw,(u_char*) &newip);
}    

/**
 * send modified ARP packets for Prophet Allocation
 */
int send_ARP(u_int16_t opcode,
             u_char* src_hw, u_char* src_ip,
             u_char* dst_hw, u_char* dst_ip){
    int c;
    u_long i;
    libnet_ptag_t t;
    u_char* payload = NULL;
    int payloadsize = 0;

    //sent state container on Reply
    if(opcode == ARP_PROPHET_REP){
        payload     = (u_char*) &_prophet;
        payloadsize = sizeof(struct prophet_states);
    }

    //build ARP header
    t = libnet_build_arp(
                        ARPHRD_ETHER,       // hardware type
                        ETHERTYPE_IP,       // protocol type
                        6,                  // hardware addr size
                        4,                  // protocol addr size
                        opcode,             // operation type
                        src_hw,             // sender hardware addr
                        src_ip,             // sender protocol addr
                        dst_hw,             // target hardware addr
                        dst_ip,             // target protocol addr
                        payload,            // payload
                        payloadsize,        // payload size
                        _libnet,            // libnet handle
                        0);                 // libnet id
    if (t == -1){
        fprintf(stderr, "Can't build ARP header: %s\n",
                libnet_geterror(_libnet));
        exit(-1);
    }

    t = libnet_autobuild_ethernet(
                        dst_hw,             // ethernet destination
                        ETHERTYPE_ARP,      // protocol type
                        _libnet);           // libnet handle
    if (t == -1){
        fprintf(stderr, "Can't build ethernet header: %s\n",
                libnet_geterror(_libnet));
        exit(-1);
    }
 
    // Write it to the wire.
    c = libnet_write(_libnet);
    if (c == -1) {
        fprintf(stderr, "Write error: %s\n",
                libnet_geterror(_libnet));
        exit(-1);
    }else{
        fprintf(stderr, "Sent %d byte ARP packet.\n", c);
    }

    libnet_clear_packet(_libnet);
    return TRUE;
}

